/* 
 * Copyright (c) 2017, Intel Corporation.
 * Intel, the Intel logo, Intel, MegaCore, NIOS II, Quartus and TalkBack 
 * words and logos are trademarks of Intel Corporation or its subsidiaries 
 * in the U.S. and/or other countries. Other marks and brands may be 
 * claimed as the property of others.   See Trademarks on intel.com for 
 * full list of Intel trademarks or the Trademarks & Brands Names Database 
 * (if Intel) or See www.Intel.com/legal (if Altera).
 * All rights reserved
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * BSD 3-Clause license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 *      - Neither Intel nor the names of its contributors may be 
 *        used to endorse or promote products derived from this 
 *        software without specific prior written permission.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/************************************************************************************
 * Copyright (C) 2018 ALTIMA Company MACNICA, Inc. All Rights Reserved.\n
 *   This software is licensed "AS IS". 
 *   Please perform use of this software by a user's own responsibility and expense.
 *   It cannot guarantee in the maker side about the damage which occurred by the ab-
 *   ility not to use or use this software, and all damage that occurred secondarily.
 ***********************************************************************************/


#include "magnes_pci.h"

#include <asm/siginfo.h>    //siginfo

MODULE_AUTHOR  ("ALTIMA Company, A Macnica Division Company");
MODULE_DESCRIPTION ("Driver for Magnes Arria 10 FPGA Board");
MODULE_SUPPORTED_DEVICE ("Mpression Magnes Board");
MODULE_LICENSE("Dual BSD/GPL");


static int magnes_init(void);
static void magnes_exit(void);
static int magnes_probe(struct pci_dev *dev, const struct pci_device_id *id);
static void magnes_remove(struct pci_dev *dev);
static int magnes_cdev_init(struct magnes_pci_dev *magnes_pci);
static void magnes_cdev_exit(struct magnes_pci_dev *magnes_pci);
static void pci_config_dump(struct magnes_pci_dev *magnes_pci);


#define DRIVER_NAME "magnes"

static const unsigned int MINOR_BASE = 0;
static const unsigned int MINOR_NUM  = 2;


static unsigned int magnes_major;
//static struct cdev magnes_cdev;
//static struct class *magnes_class = NULL;


struct file_operations s_magnes_fops = {
		.open             = magnes_open,
		.release          = magnes_close,
		.read             = magnes_read,
		.write            = magnes_write,
		.unlocked_ioctl   = magnes_ioctl,
		.mmap             = magnes_mmap,
};


static struct pci_device_id magnes_pci_ids[] = {
		{ .vendor = MAGNES_VENDOR_ID, .device = PCI_ANY_ID, \
				.subvendor = PCI_ANY_ID, .subdevice = PCI_ANY_ID},
				{ 0 },
};

MODULE_DEVICE_TABLE(pci, magnes_pci_ids);

static struct pci_driver magnes_pci_driver = {
		.name     = DRIVER_NAME,
		.id_table = magnes_pci_ids,
		.probe    = magnes_probe,
		.remove   = magnes_remove,
		/* resume, suspend are optional */
};

static void pci_config_dump(struct magnes_pci_dev *magnes_pci)
{
#if 1 //TODO: ALT Debug Start
	struct pci_dev *dev = magnes_pci->pci_dev;
	u8  temp_u8;
	u16 temp_u16;
	u32 temp_u32;

	printk("\n");
	pci_read_config_word(dev,  PCI_VENDOR_ID,           &temp_u16); printk("Vender ID           = 0x%04x \n", temp_u16);
	pci_read_config_word(dev,  PCI_DEVICE_ID,           &temp_u16); printk("Device ID           = 0x%04x \n", temp_u16);
	pci_read_config_word(dev,  PCI_COMMAND,             &temp_u16); printk("CMD                 = 0x%04x \n", temp_u16);
	pci_read_config_word(dev,  PCI_STATUS,              &temp_u16); printk("STATUS              = 0x%04x \n", temp_u16);
	pci_read_config_byte(dev,  PCI_REVISION_ID,         &temp_u8);  printk("Rev ID              = 0x%02x \n", temp_u8);
	pci_read_config_dword(dev, PCI_CLASS_REVISION,      &temp_u32); printk("Class Code          = 0x%06x \n", (temp_u32 & 0x00FFFFFF) );
	pci_read_config_byte(dev,  PCI_CACHE_LINE_SIZE,     &temp_u8);  printk("Cache Line Size     = 0x%02x \n", temp_u8);
	pci_read_config_byte(dev,  PCI_LATENCY_TIMER,       &temp_u8);  printk("Mst Latency Timer   = 0x%02x \n", temp_u8);
	pci_read_config_byte(dev,  PCI_HEADER_TYPE,         &temp_u8);  printk("Header Type         = 0x%02x \n", temp_u8);
	pci_read_config_byte(dev,  PCI_BIST,                &temp_u8);  printk("BIST                = 0x%02x \n", temp_u8);
	pci_read_config_dword(dev, PCI_BASE_ADDRESS_0,      &temp_u32); printk("BAR0                = 0x%08x \n", temp_u32);
	magnes_pci->bar_phys_addr[0] = temp_u32;
	pci_read_config_dword(dev, PCI_BASE_ADDRESS_1,      &temp_u32); printk("BAR1                = 0x%08x \n", temp_u32);
	magnes_pci->bar_phys_addr[1] = temp_u32;
	pci_read_config_dword(dev, PCI_BASE_ADDRESS_2,      &temp_u32); printk("BAR2                = 0x%08x \n", temp_u32);
	magnes_pci->bar_phys_addr[2] = temp_u32;
	pci_read_config_dword(dev, PCI_BASE_ADDRESS_3,      &temp_u32); printk("BAR3                = 0x%08x \n", temp_u32);
	magnes_pci->bar_phys_addr[3] = temp_u32;
	pci_read_config_dword(dev, PCI_BASE_ADDRESS_4,      &temp_u32); printk("BAR4                = 0x%08x \n", temp_u32);
	magnes_pci->bar_phys_addr[4] = temp_u32;
	pci_read_config_dword(dev, PCI_BASE_ADDRESS_5,      &temp_u32); printk("BAR5                = 0x%08x \n", temp_u32);
	magnes_pci->bar_phys_addr[5] = temp_u32;
	pci_read_config_dword(dev, PCI_CARDBUS_CIS,         &temp_u32); printk("Cardbus CIS Pointer = 0x%08x \n", temp_u32);
	pci_read_config_word(dev,  PCI_SUBSYSTEM_VENDOR_ID, &temp_u16); printk("SubSys Vender ID    = 0x%04x \n", temp_u16);
	pci_read_config_word(dev,  PCI_SUBSYSTEM_ID,        &temp_u16); printk("SubSys  ID          = 0x%04x \n", temp_u16);
	pci_read_config_dword(dev, PCI_ROM_ADDRESS,         &temp_u32); printk("Expansion LOM Base  = 0x%08x \n", temp_u32);
	pci_read_config_byte(dev,  PCI_CAPABILITY_LIST,     &temp_u8);  printk("Capability List     = 0x%02x \n", temp_u8);
	pci_read_config_byte(dev,  PCI_INTERRUPT_LINE,      &temp_u8);  printk("Int Line            = 0x%02x \n", temp_u8);
	pci_read_config_byte(dev,  PCI_INTERRUPT_PIN,       &temp_u8);  printk("Int Pin             = 0x%02x \n", temp_u8);
	pci_read_config_byte(dev,  PCI_MIN_GNT,             &temp_u8);  printk("Min Gnt             = 0x%02x \n", temp_u8);
	pci_read_config_byte(dev,  PCI_MAX_LAT,             &temp_u8);  printk("Max Latency         = 0x%02x \n", temp_u8);

	printk("\n");
#endif //ALT Debug End
}

static int scan_bars(struct magnes_pci_dev *magnes_pci, struct pci_dev *dev)
{
	int i;
	for (i = 0; i < MAGNES_PCI_NUM_BARS; i++) {
		unsigned long bar_start = pci_resource_start(dev, i);
		if (bar_start) {
			unsigned long bar_end = pci_resource_end(dev, i);
			unsigned long bar_flags = pci_resource_flags(dev, i);
			printk("BAR[%d] 0x%08lx-0x%08lx flags 0x%08lx \n",
					i, bar_start, bar_end, bar_flags);
		}
	}
	return 0;
}


static int map_bars(struct magnes_pci_dev *magnes_pci, struct pci_dev *dev)
{
	int i;
	for (i = 0; i < MAGNES_PCI_NUM_BARS; i++){
		unsigned long bar_start = pci_resource_start(dev, i);
		unsigned long bar_end = pci_resource_end(dev, i);
		unsigned long bar_length = bar_end - bar_start + 1;
		magnes_pci->bar_length[i] = bar_length;

		if (!bar_start || !bar_end) {
			magnes_pci->bar_length[i] = 0;
			continue;
		}

		if (bar_length < 1) {
			printk("BAR #%d length is less than 1 byte \n", i);
			continue;
		}

		/* map the device memory or IO region into kernel virtual
		 * address space */
		//magnes_pci->bar[i] = ioremap (bar_start, bar_length);
		magnes_pci->bar[i] = ioremap_nocache (bar_start, bar_length);


		if (!magnes_pci->bar[i]) {
			printk("Could not map BAR #%d. \n", i);
			return -1;
		}

		printk("BAR[%d] mapped at 0x%p with length %lu. \n", i,
				magnes_pci->bar[i], bar_length);
	}
	return 0;
}

static void free_bars(struct magnes_pci_dev *magnes_pci, struct pci_dev *dev) {

	int i;
	for (i = 0; i < MAGNES_PCI_NUM_BARS; i++) {
		if (magnes_pci->bar[i]) {
			pci_iounmap(dev, magnes_pci->bar[i]);
			magnes_pci->bar[i] = NULL;
		}
	}
}

static struct pci_dev* find_upstream_dev (struct pci_dev *dev) {
	struct pci_bus *bus = 0;
	struct pci_dev *bridge = 0;
	struct pci_dev *cur = 0;
	int found_dev = 0;

	bus = dev->bus;
	if (bus == 0) {
		printk("Device doesn't have an associated bus! \n");
		return 0;
	}

	bridge = bus->self;
	if (bridge == 0) {
		printk("Can't get the bridge for the bus! \n");
		return 0;
	}

	printk("Upstream device %x/%x, bus:slot.func %02x:%02x.%02x \n",
			bridge->vendor, bridge->device,
			bridge->bus->number, PCI_SLOT(bridge->devfn), PCI_FUNC(bridge->devfn));

	printk("List of downstream devices:");
	list_for_each_entry (cur, &bus->devices, bus_list) {
		if (cur != 0) {
			printk("  %x/%x \n", cur->vendor, cur->device);
			if (cur == dev) {
				found_dev = 1;
			}
		}
	}

	if (found_dev) {
		return bridge;
	} else {
		printk("Couldn't find upstream device! \n");
		return 0;
	}
}

void retrain_gen2 (struct magnes_pci_dev *magnes_pci) {

	struct pci_dev *dev = magnes_pci->pci_dev;
	u16 linkstat, speed, width;
	struct pci_dev *upstream;
	int pos, upos;
	u16 status_reg, control_reg, link_cap_reg;
	u16 status, control;
	u32 link_cap;
	int training, timeout;

	/* Defines for some special PCIe control bits */
#define DISABLE_LINK_BIT         (1 << 4)
#define RETRAIN_LINK_BIT         (1 << 5)
#define TRAINING_IN_PROGRESS_BIT (1 << 11)
#define LINKSPEED_2_5_GB         (0x1)
#define LINKSPEED_5_0_GB         (0x2)

	pos = pci_find_capability (dev, PCI_CAP_ID_EXP);
	if (!pos) {
		printk("Can't find PCI Express capability! \n");
		return;
	}

	/* Find root node for this bus and tell it to retrain itself. */
	upstream = magnes_pci->upstream;
	if (upstream == NULL) {
		return;
	}
	upos = pci_find_capability (upstream, PCI_CAP_ID_EXP);
	status_reg = upos + PCI_EXP_LNKSTA;
	control_reg = upos + PCI_EXP_LNKCTL;
	link_cap_reg = upos + PCI_EXP_LNKCAP;
	pci_read_config_word (upstream, status_reg, &status);
	pci_read_config_word (upstream, control_reg, &control);
	pci_read_config_dword (upstream, link_cap_reg, &link_cap);

	pci_read_config_word (dev, pos + PCI_EXP_LNKSTA, &linkstat);
	pci_read_config_dword (upstream, link_cap_reg, &link_cap);
	speed = linkstat & PCI_EXP_LNKSTA_CLS;
	width = (linkstat & PCI_EXP_LNKSTA_NLW) >> PCI_EXP_LNKSTA_NLW_SHIFT;

	if (speed == LINKSPEED_2_5_GB) {
		printk("Link is operating at 2.5 GT/s with %d lanes. Need to retrain. \n", width);
	} else if (speed == LINKSPEED_5_0_GB) {
		printk("Link is operating at 5.0 GT/s with %d lanes. \n", width);
		if (width == 8) {
			printk("  All is good! \n");
			return;
		} else {
			printk("  Need to retrain. \n");
		}
	} else {
		printk("Not sure what's going on. Retraining. \n");
	}

	/* Perform the training. */
	training = 1;
	timeout = 0;
	pci_read_config_word (upstream, control_reg, &control);
	pci_write_config_word (upstream, control_reg, control | RETRAIN_LINK_BIT);

	while (training && timeout < 50)
	{
		pci_read_config_word (upstream, status_reg, &status);
		training = (status & TRAINING_IN_PROGRESS_BIT);
		msleep (1); /* 1 ms */
		++timeout;
	}
	if(training)
	{
		printk("Error: Link training timed out. \n");
		printk("PCIe link not established. \n");
	}
	else
	{
		printk("Link training completed in %d ms. \n", timeout);
	}


	/* Verify that it's a 5 GT/s link now */
	pci_read_config_word (dev, pos + PCI_EXP_LNKSTA, &linkstat);
	pci_read_config_dword (upstream, link_cap_reg, &link_cap);
	speed = linkstat & PCI_EXP_LNKSTA_CLS;
	width = (linkstat & PCI_EXP_LNKSTA_NLW) >> PCI_EXP_LNKSTA_NLW_SHIFT;

	if(speed == LINKSPEED_5_0_GB)
	{
		printk("Link operating at 5 GT/s with %d lanes \n", width);
	}
	else
	{
		printk("** WARNING: Link training failed.  Link operating at 2.5 GT/s with %d lanes.\n", width);
	}

	return;
}

static int magnes_probe(struct pci_dev *dev, const struct pci_device_id *id)
{

	int retval;
	struct magnes_pci_dev *magnes_pci = 0;

	printk(" probe (dev = 0x%p, pciid = 0x%p)\n", dev, id);
	printk(" vendor = 0x%x \n", dev->vendor);
	printk(" device = 0x%x \n", dev->device);
	printk(" class  = 0x%x \n", dev->class);
	printk(" bus:slot.func = %02x:%02x.%02x \n", dev->bus->number, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn));
	printk("\n");


	magnes_pci = kzalloc(sizeof(struct magnes_pci_dev), GFP_KERNEL);
	if (!magnes_pci) {
		printk("Couldn't allocate memory!\n");
		goto fail_kzalloc;
	}

	magnes_pci->pci_dev = dev;
	dev_set_drvdata(&dev->dev, (void*)magnes_pci);
	magnes_pci->upstream = find_upstream_dev (dev);

	retrain_gen2 (magnes_pci);

	if (pci_enable_device(dev)) {
		printk("pci_enable_device() failed");
		goto fail_enable;
	}

	pci_set_master(dev);

	if (pci_request_regions(dev, DRIVER_NAME)) {
		goto fail_regions;
	}

	scan_bars(magnes_pci, dev);
	if (map_bars(magnes_pci, dev)) {
		goto fail_map_bars;
	}

	pci_config_dump(magnes_pci);

	retval = magnes_cdev_init(magnes_pci);
	if (retval != 0) {
		printk("magnes_pci: can't create class\n");
		goto err_cdev_create;
	}

	magnes_pci->mmap_target = 0;

	//TODO: Test for BAR Access
	unsigned int * target_bar;
	target_bar = (unsigned int *)magnes_pci->bar[2];
	target_bar[0x4000 >> 2] = 0xf;


	//magnes_pci->cdev = &magnes_cdev;

	return 0;


	/* ERROR HANDLING */
	fail_map_bars:
	pci_release_regions(dev);
	pci_disable_device (dev);

	fail_regions:
	fail_enable:
	magnes_cdev_exit(magnes_pci);
	err_cdev_create:
	kfree (magnes_pci);

	fail_kzalloc:
	return -1;
}

static void magnes_remove(struct pci_dev *dev)
{
	struct magnes_pci_dev *magnes_pci = 0;
	printk(": dev is %p", dev);

	if (dev == 0) {
		printk(": dev is 0");
		return;
	}

	magnes_pci = (struct magnes_pci_dev*) dev_get_drvdata(&dev->dev);
	if (magnes_pci == 0) {
		printk(": magnes_pci_dev is 0");
		return;
	}

	magnes_cdev_exit(magnes_pci);
	free_bars (magnes_pci, dev);
	pci_disable_device(dev);
	pci_release_regions(dev);

	kfree (magnes_pci);
}

static int magnes_cdev_init(struct magnes_pci_dev *magnes_pci)
{
	printk("magnes_cdev_init\n");

	int alloc_ret = 0;
	int cdev_err = 0;
	dev_t dev;

	alloc_ret = alloc_chrdev_region(&dev, MINOR_BASE, MINOR_NUM, DRIVER_NAME);
	if (alloc_ret != 0) {
		printk(KERN_ERR  "alloc_chrdev_region = %d\n", alloc_ret);
		return -1;
	}

	magnes_major = MAJOR(dev);
	dev = MKDEV(magnes_major, MINOR_BASE);

	cdev_init(&(magnes_pci->cdev), &s_magnes_fops);
	magnes_pci->cdev.owner = THIS_MODULE;

	cdev_err = cdev_add(&(magnes_pci->cdev), dev, MINOR_NUM);
	if (cdev_err != 0) {
		printk(KERN_ERR  "cdev_add = %d\n", alloc_ret);
		unregister_chrdev_region(dev, MINOR_NUM);
		return -1;
	}

	magnes_pci->magnes_class = class_create(THIS_MODULE, "magnes");
	if (IS_ERR(magnes_pci->magnes_class)) {
		printk(KERN_ERR  "class_create\n");
		cdev_del(&(magnes_pci->cdev));
		unregister_chrdev_region(dev, MINOR_NUM);
		return -1;
	}

	for (int minor = MINOR_BASE; minor < MINOR_BASE + MINOR_NUM; minor++) {
		device_create(magnes_pci->magnes_class, NULL, MKDEV(magnes_major, minor), NULL, "magnes%d", minor);
	}

	return 0;
}

static void magnes_cdev_exit(struct magnes_pci_dev *magnes_pci)
{
	printk("magnes_cdev_exit\n");

	dev_t dev = MKDEV(magnes_major, MINOR_BASE);

	for (int minor = MINOR_BASE; minor < MINOR_BASE + MINOR_NUM; minor++) {
		device_destroy(magnes_pci->magnes_class, MKDEV(magnes_major, minor));
	}
	class_destroy(magnes_pci->magnes_class);
	cdev_del(&(magnes_pci->cdev));

	unregister_chrdev_region(dev, MINOR_NUM);
}

static int magnes_init(void)
{
	int retval;

	printk("---------------------------------------------------------------\n");
	printk("---------------------------------------------------------------\n");
	printk(" __       __                                                   \n");
	printk("/  $     /  |                                                  \n");
	printk("$$  $   /$$ |  ______    ______   _______    ______    _______ \n");
	printk("$$$  $ /$$$ | /      $  /      $ /       $  /      $  /       |\n");
	printk("$$$$  /$$$$ | $$$$$$  |/$$$$$$  |$$$$$$$  |/$$$$$$  |/$$$$$$$/ \n");
	printk("$$ $$ $$/$$ | /    $$ |$$ |  $$ |$$ |  $$ |$$    $$ |$$      $ \n");
	printk("$$ |$$$/ $$ |/$$$$$$$ |$$ $__$$ |$$ |  $$ |$$$$$$$$/  $$$$$$  |\n");
	printk("$$ | $/  $$ |$$    $$ |$$    $$ |$$ |  $$ |$$       |/     $$/ \n");
	printk("$$/      $$/  $$$$$$$/  $$$$$$$ |$$/   $$/  $$$$$$$/ $$$$$$$/  \n");
	printk("                       /  $__$$ |                              \n");
	printk("                       $$    $$/                               \n");
	printk("                        $$$$$$/                                \n");
	printk("---------------------------------------------------------------\n");
	printk("---------------------------------------------------------------\n");

	printk("magnes_init\n");


	/* register this driver with the PCI bus driver */
	printk("pci_register_driver");
	retval =  pci_register_driver(&magnes_pci_driver);
	if (retval) {
		printk("magnes_pci: can't register pci driver\n");
		goto err_cdev_create;
	}
	printk("success");
	return 0;

	/* error handling */
	err_cdev_create:
	//magnes_cdev_exit();

	return retval;
}

static void magnes_exit(void)
{
	printk("magnes_exit\n");

	pci_unregister_driver(&magnes_pci_driver);
	//magnes_cdev_exit();

}

module_init(magnes_init);
module_exit(magnes_exit);
